///////////////////////////////////////////////////////////////////////////////
// GiDML_IO : a GiD Mesh Llbrary (GiDML) module
//
// The input/output module for the GiD Mesh Library modules
//
// Copyright (C) 2017
// International Center for Numerical Methods in Engineering (CIMNE)
// C/Gran Capita s/n, North Campus UPC, C-1 building, 08034 Barcelona, SPAIN.
// www.cimne.com
// GiD Develoment Team, pre and post-processing department.
// www.gidhome.com/gidml
// gidml@cimne.upc.edu
//
// GiDML_IO GiDML module is propietary of CIMNE. It is public and free, so
// it can be used without restriction and can be integrated into any
// other software.
// It can be downloaded from:
// https://www.gidhome.com/gid-plus/gidml-modules/11803/gidml_io/
// For more information, please, contact gidml@cimne.upc.edu or visit our website:
// www.gidhome.com/gidml
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
// This GiDML module contains all the functions to get and set all the needed
// data from the GiDInput and GiDOuput handles, needed to interact with the
// different GiDML modules.
// Also some global declarations and definitions (#typedef, #define, ...) common
// to all the GiDML modules are present in it.
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
// gidml_io.h
//
// Header file of the gidml_io library. It is also the user-level header file.
/////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <stdlib.h>

#define GiDML_IO_VERSION "1.10.0"
#define GiDML_IO_VERSION_MAJOR 1
#define GiDML_IO_VERSION_MINOR 10
#define GiDML_IO_VERSION_PATCH 0

#define GiDML_IO_FORMAT_VERSION "1.4"

typedef void *GiDInput_Handle;
typedef void *GiDOutput_Handle;
typedef void *GiDIO_Handle;


enum ElemType {
  GID_NONE_ELEMENT,
  GID_LINE_ELEMENT,
  GID_TRIANGLE_ELEMENT,
  GID_QUADRILATERAL_ELEMENT,
  GID_TETRAHEDRON_ELEMENT,
  GID_HEXAHEDRON_ELEMENT,
  GID_PRISM_ELEMENT,
  GID_POINT_ELEMENT,
  GID_PYRAMID_ELEMENT,
  GID_SPHERE_ELEMENT,
  GID_CIRCLE_ELEMENT
};

static const char *ElemTypeNames[] = { "None",       "Linear",    "Triangle", "Quadrilateral",
                                       "Tetrahedra", "Hexahedra", "Prism",    "Point",
                                       "Pyramid",    "Sphere",    "Circle" };
inline int GetNumElemTypes() { return ( int )( sizeof( ElemTypeNames ) / sizeof( char * ) ); }
inline const char *GetElementTypeName( ElemType etype ) { return ElemTypeNames[etype]; }


enum GIDML_TYPE_OF_ENTITY {
  GIDML_NODE,
  GIDML_EDGE,
  GIDML_FACE,
  GIDML_ELEMENT,
  GIDML_UNKNOWN
};
enum GIDML_TYPE_OF_ATTRIBUTE {
  GIDML_USER_ATTRIBUTE,
  GIDML_MODULE_ATTRIBUTE
};
enum GIDML_TYPE_OF_VALUE {
  GIDML_TYPE_INTEGER,
  GIDML_TYPE_DOUBLE,
  GIDML_TYPE_UNKNOWN
};

// FUNCTIONS TO GET AND SET PARAMETERS FROM THE GiDInput and GiDOutput Handles

//**********************************************************************
// GiDML Module information and data creator. Common functions for GiDInputHandle and
// GiDOutputHandle (GiDIO_Handle)
//**********************************************************************

// GiDML IO module information
const char *GiDML_IO_GetVersion();
const char *GiDML_IO_GetFormatVersion();
double GiDML_IO_GetFormatVersion_Number( const char *version );
const char *GIDML_IO_GetMinimumFormatVersionToSave( const GiDIO_Handle hdl );

// GiD mesh library module information
const char *GiDML_IO_GetGiDMeshLibraryModuleName( const GiDIO_Handle hdl );
const char *GiDML_IO_GetGiDMeshLibraryModuleFormatVersion( const GiDIO_Handle hdl );
void GiDML_IO_SetGiDMeshLibraryModuleName( const char *module_name, const GiDIO_Handle hdl );
void GiDML_IO_SetGiDMeshLibraryModuleFormatVersion( const char *module_format_version,
                                                    const GiDIO_Handle hdl );

// Data creator
void GiDML_IO_GetGiDMeshLibraryIOCreator( const GiDIO_Handle hdl, char *io_creator );
void GiDML_IO_SetGiDMeshLibraryIOCreator( const char *gidml_io_creator, const GiDIO_Handle hdl );


//**********************************************************************
// GiDInputHandle create and delete functions
//**********************************************************************
// Functions to get and delete GiDInput handle
GiDInput_Handle GiDML_IO_NewGiDInputHandle();
GiDInput_Handle GiDML_IO_NewGiDInputHandle( const GiDInput_Handle hdl_gin );
GiDInput_Handle GiDML_IO_NewGiDInputHandle( const char *module_name,
                                            const char *module_format_version,
                                            const char *io_creator );
GiDInput_Handle GiDML_IO_NewGiDInputHandle( const char *module_name,
                                            const char *module_format_version,
                                            const char *io_creator, const int ndime );

GiDInput_Handle GiDML_IO_NewGiDInputHandleWithNodes( const char *module_name,
                                                     const char *module_format_version,
                                                     const char *io_creator, const int ndime,
                                                     const int num_nodes, const double *coords );

GiDInput_Handle GiDML_IO_NewGiDInputHandleWithFaces(
    const char *module_name, const char *module_format_version, const char *io_creator,
    const int ndime, const int num_nodes, const double *coords, const int num_faces, int *faces,
    const ElemType type_faces_mesh, const int nnode );

GiDInput_Handle GiDML_IO_NewGiDInputHandleWithVolumeElements(
    const char *module_name, const char *module_format_version, const char *io_creator,
    const int ndime, const int num_nodes, const double *coords, const int num_elements, int *elems,
    const ElemType mesh_elem_type, const int nnode );

void GiDML_IO_DeleteGiDInputHandle( GiDInput_Handle &hdl_gin );

// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDINPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteGiDInputContent( GiDInput_Handle &hdl_gin );


//**********************************************************************
// Callback functions
//**********************************************************************
// callback to be raised to provide the main program information of current status
//(e.g. to show advance bar and cancel process, or print messages in console
// ( void *) should be avoided for pointer to functions
// http://stackoverflow.com/questions/14543801/iso-c-forbids-casting-between-pointer-to-function-and-pointer-to-object
// http://stackoverflow.com/questions/16026183/gcc-warns-about-function-pointer-to-object-pointer-cast
// http://stackoverflow.com/questions/1096341/function-pointers-casting-in-c
// this should be used instead:
// typedef void (*gidml_fnc_ptr)();
// void SetCallbackStatus(GiDInput_Handle& hdl_gin,gidml_fnc_ptr fn,void* client_data);
// also for the SetCallbackStatus & co. in the derivate libraries like gidml_octree_mesher.h
void GiDML_IO_SetCallbackStatus( GiDInput_Handle &hdl_gin, void *fn, void *client_data );
void GiDML_IO_GetCallbackStatus( const GiDInput_Handle &hdl_gin, void *&fn, void *&client_data );

void GiDML_IO_SetCallbackError( GiDInput_Handle &hdl_gin, void *fn, void *client_data );
void GiDML_IO_GetCallbackError( const GiDInput_Handle &hdl_gin, void *&fn, void *&client_data );


//**********************************************************************
// GiDOutputHandle create and delete functions
//**********************************************************************
// Functions to get and delete GiDOutput handle
GiDOutput_Handle GiDML_IO_NewGiDOutputHandle( const GiDOutput_Handle hdl_gout );
GiDOutput_Handle GiDML_IO_NewGiDOutputHandle();
void GiDML_IO_DeleteGiDOutputHandle( GiDOutput_Handle &hdl_gout );

// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDOUPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteGiDOutputContent( GiDOutput_Handle hdl_gout );

//**********************************************************************
// Mesh definition functions. Common Functions for GiDInputHandle and GiDOutputHandle (GiDIO_Handle)
//**********************************************************************

// Mesh dimension
int GiDML_IO_GetMeshDimension( const GiDIO_Handle hdl );
void GiDML_IO_SetMeshDimension( const GiDIO_Handle hdl, const int dimension );


// Nodes
void GiDML_IO_SetNodesCoords( GiDIO_Handle hdl, const int num_nodes, const double *coords );
int GiDML_IO_GetNodesCoords( const GiDIO_Handle hdl, double *&nodes_coords);
int GiDML_IO_GetNumberOfNodes( const GiDIO_Handle hdl );
void GiDML_IO_SetNumberOfNodes( GiDIO_Handle hdl, const int num_nodes );
void GiDML_IO_DeleteNodes( GiDIO_Handle hdl );
// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDOUPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteNodesContent( GiDIO_Handle &hdl );

// Edges
void GiDML_IO_SetEdges( GiDIO_Handle hdl, int num_edges, int *connectivities,
                        const int n_elem_types, ElemType *etypes, int *nnodes,
                        int *n_elems_by_type );
void GiDML_IO_SetEdges( GiDIO_Handle hdl, const int num_edges, int *connectivities,
                        const ElemType etype, const int nnode );
int GiDML_IO_GetEdges( GiDIO_Handle hdl, int *&connec, ElemType *&etypes, int *&nnodes,
                       int *&n_elems_by_type );
int GiDML_IO_GetEdges( const GiDIO_Handle hdl, int *&connec );
int GiDML_IO_GetNumberOfEdges( const GiDIO_Handle hdl );
int GiDML_IO_GetNumberOfElementTypesInEdges( const GiDIO_Handle hdl );
void GiDML_IO_DeleteEdges( GiDIO_Handle hdl );
// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDOUPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteEdgesContent( GiDIO_Handle &hdl );


// Faces
void GiDML_IO_SetFaces( GiDIO_Handle hdl, const int num_faces, int *connectivities,
                        const int n_elem_types, ElemType *etypes, int *nnodes,
                        int *n_elems_by_type );
void GiDML_IO_SetFaces( GiDIO_Handle hdl, const int num_faces, int *connectivities,
                        const ElemType etype, const int nnode );
int GiDML_IO_GetFaces( const GiDIO_Handle hdl, int *&connec, ElemType *&etypes, int *&nnodes,
                       int *&n_elems_by_type );
int GiDML_IO_GetFaces( const GiDIO_Handle hdl, int *&connec );
int GiDML_IO_GetNumberOfFaces( const GiDIO_Handle hdl );
int GiDML_IO_GetNumberOfElementTypesInFaces( const GiDIO_Handle hdl );
void GiDML_IO_DeleteFaces( GiDIO_Handle hdl );
// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDOUPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteFacesContent( GiDIO_Handle &hdl );


// Elements
void GiDML_IO_SetElements( GiDIO_Handle hdl, int num_elems, int *connectivities,
                           const int n_elem_types, ElemType *etypes, int *nnodes,
                           int *n_elems_by_type );
void GiDML_IO_SetElements( GiDIO_Handle hdl, const int num_elements, int *connectivities,
                           const ElemType etype, const int nnode );
int GiDML_IO_GetElements( const GiDIO_Handle hdl, int *&connec, ElemType *&etypes, int *&nnodes,
                          int *&n_elems_by_type );
int GiDML_IO_GetElements( const GiDIO_Handle hdl, int *&connec );
int GiDML_IO_GetNumberOfElements( const GiDIO_Handle hdl );
int GiDML_IO_GetNumberOfElementTypesInElements( const GiDIO_Handle hdl );
void GiDML_IO_DeleteElements( GiDIO_Handle hdl );
// DANGER FUNCTION: ONLY TO BE USED IF THE CONTENT OF GIDOUPUT CONTENT IS CREATED USING 'NEW'
void GiDML_IO_DeleteElementsContent( GiDIO_Handle &hdl );

// generic elems
void GiDML_IO_SetGenericElements( GiDIO_Handle hdl, const GIDML_TYPE_OF_ENTITY entity_type,
                                  int num_elems, int *connectivities, const int n_elem_types,
                                  ElemType *etypes, int *nnodes, int *n_elems_by_type );
void GiDML_IO_SetGenericElements( GiDIO_Handle hdl, const GIDML_TYPE_OF_ENTITY entity_type,
                                  const int num_elems, int *connectivities, const ElemType etype,
                                  const int nnodes );
int GiDML_IO_GetGenericElements( const GiDIO_Handle hdl, const GIDML_TYPE_OF_ENTITY entity_type,
                                 int *&connec, ElemType *&etypes, int *&nnodes,
                                 int *&n_elems_by_type );
int GiDML_IO_GetGenericElements( const GiDIO_Handle hdl, const GIDML_TYPE_OF_ENTITY entity_type,
                                 int *&connec );
int GiDML_IO_GetNumberOfGenericElements( const GiDIO_Handle hdl,
                                         const GIDML_TYPE_OF_ENTITY entity_type );
int GiDML_IO_GetNumberOfGenericElementTypes( const GiDIO_Handle hdl,
                                             const GIDML_TYPE_OF_ENTITY entity_type );
ElemType GiDML_IO_GetGenericElementType( const GiDIO_Handle hdl,
                                         const GIDML_TYPE_OF_ENTITY entity_type, const int ipos );
int GiDML_IO_GetNNodeGenericElement( const GiDIO_Handle hdl, const GIDML_TYPE_OF_ENTITY entity_type,
                                     const int ipos );


//**********************************************************************
// Additional parameters. Common Functions for GiDInputHandle and GiDOutputHandle (GiDIO_Handle)
//**********************************************************************

// Attributes
void *GiDML_IO_CreateAttribute( GiDIO_Handle hdl, const char *attribute_name,
                                const GIDML_TYPE_OF_ENTITY entity_type,
                                const GIDML_TYPE_OF_ATTRIBUTE attribute_type,
                                const GIDML_TYPE_OF_VALUE value_type );

int GiDML_IO_SetAttribute( GiDIO_Handle hdl, const char *attribute_name, void *attribute_array,
                           const GIDML_TYPE_OF_ENTITY entity_type,
                           const GIDML_TYPE_OF_ATTRIBUTE attribute_type,
                           const GIDML_TYPE_OF_VALUE value_type );
// IMPORTANT: a copy of attribute_array is made inside the function GiDML_IO_SetAttribute.

void *GiDML_IO_GetAttribute( GiDIO_Handle hdl, const char *attribute_name,
                             const GIDML_TYPE_OF_ENTITY entity_type );
GIDML_TYPE_OF_VALUE GiDML_IO_GetAttributeType( GiDIO_Handle hdl, const char *attribute_name,
                                               const GIDML_TYPE_OF_ENTITY entity_type );
int GiDML_IO_DeleteAttribute( GiDIO_Handle hdl, const char *attribute_name,
                              const GIDML_TYPE_OF_ENTITY entity_type );
int GiDML_IO_DeleteAttributesEntitiesVector( GiDIO_Handle hdl,
                                             const GIDML_TYPE_OF_ENTITY entity_type );
int GiDML_IO_GetNumberOfAttributesThatMustBeTransmitted( GiDIO_Handle hdl,
                                                         const GIDML_TYPE_OF_ENTITY entity_type );
const char *GiDML_IO_GetAttributeNameThatMustBeTransmitted( GiDIO_Handle hdl,
                                                            const GIDML_TYPE_OF_ENTITY entity_type,
                                                            const int i );

// Parameters
int GiDML_IO_SetParameter( GiDIO_Handle hdl, const char *parameter_name, double value );
int GiDML_IO_SetParameter( GiDIO_Handle hdl, const char *parameter_name, int value );
int GiDML_IO_GetParameter( const GiDIO_Handle hdl, const char *parameter_name, double &value );
bool GiDML_IO_ParameterExists( const GiDIO_Handle hdl, const char *parameter_name );
size_t GiDML_IO_GetNumberOfParameters( const GiDIO_Handle hdl );

// Arrays of parameters
int GiDML_IO_SetParameterVector( GiDIO_Handle hdl, const char *parameter_name, int *array_of_values,
                                 const int array_dimension );
int GiDML_IO_SetParameterVector( GiDIO_Handle hdl, const char *parameter_name,
                                 double *array_of_values, const int array_dimension );
int GiDML_IO_GetParameterVector( GiDIO_Handle hdl, const char *parameter_name,
                                 int *&array_of_values, int &array_dimension );
int GiDML_IO_GetParameterVector( GiDIO_Handle hdl, const char *parameter_name,
                                 double *&array_of_values, int &array_dimension );
int GiDML_IO_DeleteParameterVector( GiDIO_Handle hdl, const char *parameter_name );
size_t GiDML_IO_GetNumberOfParametersVector( const GiDIO_Handle hdl );

void GiDML_IO_DeleteParametersContent( GiDIO_Handle hdl ); // clear all,parameters and arrays of
                                                           // parameters


//************************************************************************
// Write and read GiDInput and GiDOutput info using files
//************************************************************************
// Note: the files used are has the extension of '.gidml', and are binary files in format HDF5.
// The filename must include the complete path
int GiDML_IO_WriteGiDInput( const char *filename, const GiDInput_Handle hdl );
int GiDML_IO_ReadGiDInput( const char *filename, GiDInput_Handle &hdl_gin );

int GiDML_IO_WriteGiDOutput( const char *filename, const GiDOutput_Handle hdl_gout );
int GiDML_IO_ReadGiDOutput( const char *filename, GiDOutput_Handle &hdl_gout );

int GiDML_IO_WriteGiDInputAndGiDOutput( const char *filename, const GiDInput_Handle hdl_gin,
                                        const GiDOutput_Handle hdl_gout );
int GiDML_IO_ReadGiDInputAndGiDOutput( const char *filename, GiDInput_Handle &hdl_gin,
                                       GiDOutput_Handle &hdl_gout );

int GiDML_IO_WriteMeshInGiDFormat( const char *filename, const GiDIO_Handle hdl );


//************************************************************************
// Other functions
//************************************************************************
bool GiDML_IO_GiDInputsAreEqual( const GiDInput_Handle ginput_handle_1,
                                 const GiDInput_Handle ginput_handle_2 );
bool GiDML_IO_GiDOutputsAreEqual( const GiDOutput_Handle goutput_handle_1,
                                  const GiDOutput_Handle goutput_handle_2 );
